/* jmixer.cpp - part of MUSE
 * (c)2000 Denis Roio aka jaromil
 * c++ engine for multiple mp3 realtime stream mixing
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>

#include "generic.h"
#include "audioproc.h"
#include "inchannels.h"
#include "outchannels.h"
#include "libsync/synchronous.h"

struct livein {
  short inbuf[BUF_SIZE];
  int rate;
  int channels;
  bool on;
};

class Stream_mixer: public LTI::Object {
  
private:

  short audio_buffer[BUF_SIZE]; /* interleaved buffer */
  int process_buffer[BUF_SIZE];
  short inbuf[BUF_SIZE*40];

  int dsp;
  bool fullduplex;

  struct livein live_channel;
  
  void close_soundcard();

  unsigned int get_livein();
  
public:
  Stream_mixer();

  void run();
  
  bool open_soundcard();
  bool set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps);
  void stop_icecast();
  void play_chunks();
  void resume_channel(int ch);
  void pause_channel(int ch);
  void stop_channel(int ch);
  void set_channel(int ch, int playlist_pos);
  void set_volume(int ch, float vol);
  void set_speed(int ch, float speed);
  void dump_status();
  bool set_live(bool stat);
  void set_playlist(int ch, char *pl);

  Mp3Channel mp3ch[MAX_CHANNELS];
  IcecastOut ice;

  /* this is for the libsync - registering synchronized method's ids */
  static const int id_set_channel = 1;
  static const int id_stop_channel = 2;
  static const int id_pause_channel = 3;
  static const int id_resume_channel = 4;
  static const int id_set_volume = 5;
  static const int id_set_speed = 6;
  static const int id_dump_status = 7;
  static const int id_set_live = 8;
  static const int id_set_playlist = 9;
  static const int id_set_icecast = 10;
  static const int id_stop_icecast = 11;
  static const int id_quit = 12;

  bool dspout;
  bool icecastout;
  bool quit;

  int *client;

  ~Stream_mixer();
};

Stream_mixer::Stream_mixer() {
  int c = 0;

  dsp = 0;

  fprintf(stderr,"Initializing channels ...");
  
  /* init channels */
  for(c=0;c<MAX_CHANNELS;c++) {
    mp3ch[c].inbuf = inbuf;
    mp3ch[c].dsp = &dsp;
    mp3ch[c].sample_rate = SAMPLE_RATE;
  }

  /* setup livein */
  live_channel.rate = SAMPLE_RATE;
  live_channel.channels = 2;
  live_channel.on = false;

  fprintf(stderr," done.\n");

  quit = false;

  /* Initialization as a synchronous object */
  synchronous.registerObjectName("Stream_mixer");

}

void Stream_mixer::run() {

  if(debug)
    cerr << selfName() << " running on thread " << (int)pthread_self() << "\n";

  while(!quit) {
    
    synchronous.select.accept(0, id_resume_channel);
    synchronous.select.accept(1, id_stop_channel);
    synchronous.select.accept(2, id_set_channel);
    synchronous.select.accept(3, id_pause_channel);    
    synchronous.select.accept(4, id_set_volume);
    synchronous.select.accept(5, id_set_speed);
    synchronous.select.accept(6, id_dump_status);
    synchronous.select.accept(7, id_set_live);
    synchronous.select.accept(8, id_set_playlist);
    synchronous.select.accept(9, id_set_icecast);
    synchronous.select.accept(10, id_stop_icecast);

    synchronous.select.waitUntil(11, LTI::Kernel::now()+5);
    synchronous.wait();

    play_chunks();
  }

  cerr << selfName() << " exiting " << "\n";

}

Stream_mixer::~Stream_mixer() {
  int i;

  for(i=0;i<MAX_CHANNELS;i++)
    mp3ch[i].stop();

  if(dsp!=0)
    close_soundcard();
  
  if(icecastout)
    stop_icecast();

}

bool Stream_mixer::open_soundcard()
{
    int format,tstereo,speed,caps;

    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
        perror("Error opening /dev/dsp: ");
	return(false);
    } else
      fprintf(stderr,"Found soundcard on /dev/dsp, initializing ...\n");

    format = AFMT_S16_LE;
    tstereo = 1; /* i implement only stereo dsp output unless anybody complains */
    speed = SAMPLE_RATE;

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      fprintf(stderr,"full duplex is supported, good\n");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      fprintf(stderr,"full duplex is not supported\n");
      fullduplex = false;
    }

    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      fprintf(stderr,"failed to set data format\n");
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      fprintf(stderr,"something went wrong with the stereo setting\n");
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      fprintf(stderr,"speed setting failed\n");

    fprintf(stderr,"format setted to 16bit %dHz stereo.\n",SAMPLE_RATE);
    dspout = true;

    return(true);
} /* open_soundcard */

bool Stream_mixer::set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps) {

  ice.set(ip,port,pass,mountpoint,bps);
  ice.init_encoder(bps);
  
  if(ice.start())
    icecastout = true;
  else
    icecastout = false;

  if(icecastout)
    fprintf(stderr,"Estabilished %ubps stream on %s:%u/%s\n",bps,ip,port,mountpoint);
  else
    fprintf(stderr,"Got problems streaming on server %s\n",ip);

  synchronous.reply();
  return(icecastout);
}

void Stream_mixer::stop_icecast() {
  ice.stop();
  icecastout = false;

  fprintf(stderr,"Stopped streaming to icecast server\n");

  synchronous.reply();
}

void Stream_mixer::play_chunks()
{
  int i,j,num;
  //  char temp[(BUF_SIZE<<4)/44100 + ((4*1152*16)/44100+512)];
  //  char temp[BUF_SIZE];
  num=BUF_SIZE;
  
  memset(process_buffer,0,num<<2);
  
  for(i=0;i<MAX_CHANNELS;i++) {
    if((mp3ch[i].pid)&&(mp3ch[i].on)) {
      mp3ch[i].get_audio();
      if(mp3ch[i].volume!=1.0)
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + (mp3ch[i].gotin[j] * mp3ch[i].volume));
      else
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + mp3ch[i].gotin[j]);

      /* if the mp3 reached the end */
      if(mp3ch[i].end)
	mp3ch[i].ends();

    } // if (mp3ch[i].pid&&mp3ch[i].on)
  } // for i=0 to MAX_CHANNELS

  if(live_channel.on) {
    i = get_livein();
    for(j=0;j<num;j++)
      process_buffer[j] = (int) (process_buffer[j] + live_channel.inbuf[j]);
  }
  
  /* CLIPPING */
  for(j=0;j<num;j++)
    if((process_buffer[j]<CLIP)&&(process_buffer[j]>-CLIP))
      audio_buffer[j]=(short)(process_buffer[j]);
    else if (process_buffer[j]>0)
      audio_buffer[j]=MAX_SAMPLE-(CLIP_A/(CLIP_B+process_buffer[j]));
    else
      audio_buffer[j]=-(MAX_SAMPLE-(CLIP_A/(CLIP_B-process_buffer[j])));
  /* -------- */

  /* NO CLIPPING
     for(j=0;j<num;j++)
     audio_buffer[j]=(short)process_buffer[j];
    ----------- */

  if(dspout)
    write(dsp,audio_buffer,num<<1);

  if(icecastout) {
    num = ice.encode(&audio_buffer[0],num>>1,process_buffer,BUF_SIZE);
    if(num>0)
      if(!ice.send((short int *)process_buffer,num))
	fprintf(stderr,"Send error: cannot send stream to icecast server\n");
      else if(num==-1) {
	fprintf(stderr,"Encoding failed: mp3buffer overflowed\n");
	fprintf(stderr,"UACK! bailing out!\n");
	exit(1);
      } 
  }
}

void Stream_mixer::resume_channel(int ch) {
  mp3ch[ch].on = true;
  synchronous.reply();
}

void Stream_mixer::pause_channel(int ch) {
  mp3ch[ch].on = false;
  synchronous.reply();
}

void Stream_mixer::set_channel(int ch, int playlist_pos) {
  mp3ch[ch].start(playlist_pos);

  fprintf(stderr,"executed set_channel(%u, %u) as requested\n", ch, playlist_pos);
  
  synchronous.reply();
}

void Stream_mixer::set_volume(int ch, float vol) {
  mp3ch[ch].volume = vol;

  fprintf(stderr,"executed set_volume(%u, %f) as requested\n", ch, vol);

  synchronous.reply();
}

void Stream_mixer::set_speed(int ch, float speed) {
  mp3ch[ch].speed = speed;
  synchronous.reply();
}

void Stream_mixer::set_playlist(int ch, char *pl) {
  mp3ch[ch].get_playlist(pl);
  synchronous.reply();
}

void Stream_mixer::stop_channel(int ch) {
  mp3ch[ch].stop();
  synchronous.reply();
}

void Stream_mixer::dump_status() {
  int c;
  fprintf(stderr,"__________dumping_status__________\n");
  for(c=0;c<MAX_CHANNELS;c++) {
    
    if(!mp3ch[c].on)
      fprintf(stderr,"[channel %02u OFF] : PID %u\n",c,mp3ch[c].pid);
    else {
      fprintf(stderr,"[channel %02u ON ] : PID %u\n",c,mp3ch[c].pid);
      fprintf(stderr,"\tname:\t%s\n",mp3ch[c].name);
      fprintf(stderr,"\tvolume:\t%f\n",mp3ch[c].volume);
      fprintf(stderr,"\tspeed:\t%f\n",mp3ch[c].speed);
      fprintf(stderr,"\trate:\t%u\n",mp3ch[c].rate);
      fprintf(stderr,"\tchannels:\t%u\n",mp3ch[c].channels);
      fprintf(stderr,"\tend:\t%u\n",mp3ch[c].end);
      fprintf(stderr,"\tP stream:\t%u\n",mp3ch[c].stream[0]);
      fprintf(stderr,"\tC stream:\t%u\n",mp3ch[c].stream[1]);
    }
  }
  fprintf(stderr,"__________________________________\n");    
  synchronous.reply();
}

bool Stream_mixer::set_live(bool stat) {
  if(!((dspout)&&(!fullduplex)&&(stat)))
    live_channel.on = stat;
  synchronous.reply();
  return(live_channel.on&stat);
}

unsigned int Stream_mixer::get_livein() {
  unsigned int num_samples = 0;
  int opt, i;
  
  num_samples=(int)((live_channel.rate*BUF_SIZE)/(SAMPLE_RATE<<1));
  opt = (num_samples*live_channel.channels)<<1;

  i=read(dsp,inbuf,opt);

  /* convert the audio into something useful */
  convert_audio(inbuf,live_channel.inbuf,live_channel.channels,2,num_samples,BUF_SIZE>>1);

  return(num_samples);
} /* get_livein */

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

